#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <syslog.h>

/* The delay time between scans */
#define DELAYTIME 10 

/* Where does the config file live */
#define CONFIGFILE "/etc/nannie.cfg"

struct tever {
    char *fname;
    struct stat *info;
    struct tever *next;
};
typedef struct tever f;

f* read_config();
void init_stats(f *);
char *get_type(int);
char *get_mode(int);

int facility=LOG_DAEMON;

int main (int argc, char *argv[]) {
	struct stat *tmp = (struct stat *) malloc (sizeof(struct stat));
    f *begin=NULL;
    f *p=NULL;
	int pid,oldmode,newmode,oldtype,newtype;
    errno=0;
    
    if(getuid()) {
        openlog(0,LOG_PID,facility);
        syslog(LOG_ERR, "Executed as uid %d - exiting",getuid());
        closelog();
        exit (1);
    }


    while(!begin) {
        begin=read_config();
        if(!begin) 
            sleep(DELAYTIME);
    }

	init_stats(begin);

	pid=fork();
	if(pid) 
		exit(0);

	for(;;) {
		sleep(DELAYTIME);

		for(p=begin;p;p=p->next) {
			if(lstat(p->fname,tmp)) {
				openlog(0,LOG_PID,facility);
				syslog(LOG_ERR, "%s - (%m)",p->fname);
				closelog();
				continue;
			}

			if(tmp->st_ino != p->info->st_ino) {
				openlog(0,LOG_PID,facility);
				syslog(LOG_ERR, "%s - (Inode has changed)",p->fname);
				closelog();
				p->info->st_ino = tmp->st_ino;
			}
			
			if(tmp->st_nlink != p->info->st_nlink) {
				openlog(0,LOG_PID,facility);
				syslog(LOG_ERR, "%s - (link count has changed)",p->fname);
				closelog();
				p->info->st_nlink = tmp->st_nlink;
			}

			if(tmp->st_uid != p->info->st_uid) {
				openlog(0,LOG_PID,facility);
				syslog(LOG_ERR, "%s - (UID has changed)",p->fname);
				closelog();
				p->info->st_uid = tmp->st_uid;
			}

			if(tmp->st_gid != p->info->st_gid) {
				openlog(0,LOG_PID,facility);
				syslog(LOG_ERR, "%s - (GID has changed)",p->fname);
				closelog();
				p->info->st_gid = tmp->st_gid;
			}

			if(tmp->st_size != p->info->st_size) {
				openlog(0,LOG_PID,facility);
				if(tmp->st_size > p->info->st_size) 
					syslog(LOG_ERR, "%s - (file has grown)",p->fname);
				else 
					syslog(LOG_ERR, "%s - (file has gotten smaller)",p->fname);
				closelog();
				p->info->st_size = tmp->st_size;
			}

			if(tmp->st_mtime != p->info->st_mtime) {
				openlog(0,LOG_PID,facility);
				syslog(LOG_ERR, "%s - (mtime has changed)",p->fname);
				closelog();
				p->info->st_mtime = tmp->st_mtime;
			}

			if(tmp->st_ctime != p->info->st_ctime) {
				openlog(0,LOG_PID,facility);
				syslog(LOG_ERR, "%s - (ctime has changed)",p->fname);
				closelog();
				p->info->st_ctime = tmp->st_ctime;
			}

			if(tmp->st_mode != p->info->st_mode) {
				newtype=tmp->st_mode&S_IFMT;
				oldtype=p->info->st_mode&S_IFMT;
				newmode=tmp->st_mode&~S_IFMT;
				oldmode=p->info->st_mode&~S_IFMT;

				openlog(0,LOG_PID,facility);
				if(newtype != oldtype) {
					syslog(LOG_ERR, "%s - (changed from %s to %s)",p->fname,get_type(oldtype),get_type(newtype));
				}
				if(newmode != oldmode) {
					syslog(LOG_ERR,"%s - (%s mode to %s mode)",p->fname,get_mode(oldmode),get_mode(newmode));
				}
				closelog();
				p->info->st_mode = tmp->st_mode;
			}
		}

	}

    return 0;
}

f* read_config() {
    struct stat *buf = (struct stat *) malloc (sizeof(struct stat));
    char *buffer=(char *) malloc (sizeof(char) * 1024);
    char *buffer2=(char *) malloc (sizeof(char) * 1024);
    DIR  *dp;
    struct dirent *dirp;
    FILE *fp;
    int i,c;
    f * r=NULL;
    f * p=NULL;

    if(lstat(CONFIGFILE,buf) == -1) {
        openlog(0,LOG_PID,facility);
        syslog(LOG_ERR, "%m");
        closelog();
        return (f *) 0;
    }

    if(!S_ISREG(buf->st_mode)) {
        openlog(0,LOG_PID,facility);
        syslog(LOG_ERR, "%s is not a regular file",CONFIGFILE);
        closelog();
        return (f *) 0;
    }

    if((buf->st_mode & ~S_IFMT) ^ (S_IRUSR | S_IWUSR)) {
        openlog(0,LOG_PID,facility);
        syslog(LOG_ERR, "%s is not mode 0600",CONFIGFILE);
        closelog();
        return (f *) 0;
    }    

    if(buf->st_uid) {
        openlog(0,LOG_PID,facility);
        syslog(LOG_ERR, "%s is not owned by root",CONFIGFILE);
        closelog();
        return (f *) 0;
    }

    fp=fopen(CONFIGFILE,"r");
    if(!fp && errno) {
        openlog(0,LOG_PID,facility);
        syslog(LOG_ERR, "%s - (%m)",CONFIGFILE);
        closelog();
        return (f *) 0;
    }

    for(c=1;fgets(buffer,1023,fp);c++) {
        i=strlen(buffer)-1;
        if(buffer[i] != '\n') {
            openlog(0,LOG_PID,facility);
            syslog(LOG_ERR, "%s line %d too long",CONFIGFILE,c);
            closelog();
            continue;
        } 

        buffer[i]='\0';

        if(lstat(buffer,buf) == -1) {
            openlog(0,LOG_PID,facility);
            syslog(LOG_ERR, "%s - (%m)",buffer);
            closelog();
            continue;
        }

        if(!(S_ISREG(buf->st_mode)) && 
           !(S_ISDIR(buf->st_mode)) &&
           !(S_ISLNK(buf->st_mode))) {
            openlog(0,LOG_PID,facility);
            syslog(LOG_ERR, "%s (inappropriate file type).",buffer);
            closelog();
            continue;
        }

        if(S_ISDIR(buf->st_mode)) {
            if(!(dp=opendir(buffer))) {
                openlog(0,LOG_PID,facility);
                syslog(LOG_ERR, "Could not open directory %s",buffer);
                closelog();
                continue;
            }
            while((dirp=readdir(dp))) {
				strcpy(buffer2,buffer);
				strcat(buffer2,"/");
				strcat(buffer2,dirp->d_name);
				if(lstat(buffer2,buf) == -1) {
					openlog(0,LOG_PID,facility);
					syslog(LOG_ERR, "%s - (%m)",buffer2);
					closelog();
					continue;
				}

				if(!(S_ISREG(buf->st_mode)) &&
				   !(S_ISLNK(buf->st_mode)) &&
				   !(S_ISDIR(buf->st_mode))) {
					   openlog(0,LOG_PID,facility);
					   syslog(LOG_ERR, "%s (inappropriate file type).",buffer);
					   closelog();
					   continue;
				}

				if(!S_ISDIR(buf->st_mode)) {
					if(!r) {
						r=(f *) malloc (sizeof(f));
						p=r;
					} else {
						p->next=(f *) malloc (sizeof(f));
						p=p->next;
					}
					p->next=NULL;
					p->info=NULL;
					p->fname=(char *) malloc (sizeof(char)*(strlen(buffer2)+1));
					strcpy(p->fname,buffer2);
				}
                
            }

        } else {
            if(!r) {
                r=(f *) malloc (sizeof(f));
                p=r;
            } else {
                p->next=(f *) malloc (sizeof(f));
                p=p->next;
            }
            p->next=NULL;
            p->info=NULL;
            p->fname=(char *) malloc (sizeof(char) * (strlen(buffer)+1));
            strcpy(p->fname,buffer);
        }
    }
    return r;
}

void init_stats(f * p) {
	for(;p;p=p->next) {
		if(!p->info)
			p->info=(struct stat *) malloc(sizeof(struct stat));
		if(lstat(p->fname,p->info)) {
			openlog(0,LOG_PID,facility);
			syslog(LOG_ERR, "%s - (%m)",p->fname);
			closelog();
		}
	}
}

char *get_type(int typ) {
	char *r=(char *) malloc (sizeof(char) * 30);
	strcpy(r,"");

	switch(typ) {
		case S_IFSOCK:
			strcpy(r,"socket");
			break;
		case S_IFLNK:
			strcpy(r,"symbolic link");
			break;
		case S_IFREG:
			strcpy(r,"regular file");
			break;
		case S_IFBLK:
			strcpy(r,"block device");
			break;
		case S_IFDIR:
			strcpy(r,"directory");
			break;
		case S_IFCHR:
			strcpy(r,"character device");
			break;
		case S_IFIFO:
			strcpy(r,"FIFO");
			break;
		default:
			strcpy(r,"unknown");
	}

	return r;
}

char *get_mode(int mod) {
	char *r=(char *) malloc (sizeof(char) * 30);
	strcpy(r,"-");

	/* u+r? */
	if(mod & S_IRUSR) 
		strcat(r,"r");
	else 
		strcat(r,"-");

	/* u+w? */
	if(mod & S_IWUSR)
		strcat(r,"w");
	else
		strcat(r,"-");

	/* u+s? u+x? */
	if(mod & S_ISUID)
		strcat(r,"s");
	else if (mod & S_IXUSR)
		strcat(r,"x");
	else
		strcat(r,"-");

	/* g+r? */
	if(mod & S_IRGRP)
		strcat(r,"r");
	else
		strcat(r,"-");

	/* g+w? */
	if(mod & S_IWGRP) 
		strcat(r,"w");
	else
		strcat(r,"-");

	/* g+s? g+x? */
	if(mod & S_ISGID)
		strcat(r,"s");
	else if (mod & S_IXGRP)
		strcat(r,"x");
	else
		strcat(r,"-");

	/* o+r? */
	if(mod & S_IROTH)
		strcat(r,"r");
	else
		strcat(r,"-");

	/* o+w? */
	if(mod & S_IWOTH)
		strcat(r,"w");
	else
		strcat(r,"-");

	/* t? o+x? */
	if(mod & S_ISVTX)
		strcat(r,"t");
	else if (mod & S_IXOTH)
		strcat(r,"x");
	else
		strcat(r,"-");

	return r;
}
